﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.IO;
using BCSV_Editor;

namespace BCSV_Editor
{
    public partial class Form1 : Form
    {
        public BCSV bcsv = new BCSV();
        public Lookup lookup = new Lookup();
        public UInt32 entryIndex = new UInt32();
        public UInt32 fieldIndex = new UInt32();
        public UInt32 typeIndex = new UInt32();
        public Object data = new Object();
        public Boolean stringOffsChanged = new Boolean();
        public Boolean selectingEntryOrField = new Boolean();
        public UInt32 stringOffs = new UInt32();

        public Form1()
        {
            InitializeComponent();
        }

        private void exitToolStripMenuItem_Click(object sender, EventArgs e)
        {
            Application.Exit();
        }

        private void openToolStripMenuItem_Click(object sender, EventArgs e)
        {
            UInt32 i;

            openFD.Filter = "BCSV Files|*.bcsv|Table Files|*.tbl|BANMT Files|*.banmt|BCAM Files|*.bcam|PA Files|*.pa|All Files|*";
            openFD.ShowDialog();
            String bcsvName = "";
            bcsvName = openFD.FileName;

            if (bcsvName.Length > 0)
            {
                this.bcsv = new BCSV(bcsvName, 0);
                this.lookup = new Lookup("lookup.txt");

                entrySelect.Items.Clear();
                fieldSelect.Items.Clear();

                for (i = 0; i < this.bcsv.header.entryCount; i++)
                {
                    entrySelect.Items.Add("Entry " + Convert.ToString(i));
                }

                for (i = 0; i < this.bcsv.header.fieldCount; i++)
                {
                    String hash = Convert.ToString(this.bcsv.fieldList[i].nameHash, 16).ToUpper();
                    hash = hash.PadLeft(8, (Char)'0');

                    //MessageBox.Show(Convert.ToString(8 - hash.Length));

                    fieldSelect.Items.Add(hash + ": " + this.lookup.lookupHash(hash));
                }

                this.stringOffsChanged = false;
                this.selectingEntryOrField = true;
            }
        }

        private void saveToolStripMenuItem_Click(object sender, EventArgs e)
        {
            this.bcsv.Write();
            MessageBox.Show("File saved");
        }

        private void Form1_Load(object sender, EventArgs e)
        {

        }

        private void fieldSelect_SelectedIndexChanged(object sender, EventArgs e)
        {
            this.selectingEntryOrField = true;
            this.fieldIndex = (UInt32)fieldSelect.SelectedIndex;
            this.typeIndex = this.bcsv.entryList[this.entryIndex].dataType[this.fieldIndex];
            typeSelect.SelectedIndex = this.bcsv.entryList[this.entryIndex].dataType[this.fieldIndex];
            dataEntry.Text = Convert.ToString(this.bcsv.entryList[this.entryIndex].data[this.fieldIndex]);
            if (this.bcsv.fieldList[fieldIndex].dataType != 1 && this.bcsv.fieldList[fieldIndex].dataType != 2 && this.bcsv.fieldList[fieldIndex].dataType != 6 && this.bcsv.fieldList[fieldIndex].dataType != 7)
            {
                try
                {
                    dataEntryShifted.Text = Convert.ToString(((Int32)this.bcsv.entryList[this.entryIndex].data[this.fieldIndex] & this.bcsv.fieldList[fieldIndex].mask) >> this.bcsv.fieldList[fieldIndex].shiftVal);
                }
                catch (InvalidCastException)
                {
                    dataEntryShifted.Text = String.Empty;
                }
            }
            else
            {
                dataEntryShifted.Text = String.Empty;
            }

            //this.selectingEntryOrField = false;

        }

        private void typeSelect_SelectedIndexChanged(object sender, EventArgs e)
        {
            this.typeIndex = (UInt32)typeSelect.SelectedIndex;
        }

        private void dataEntry_TextChanged(object sender, EventArgs e)
        {
            if (this.typeIndex != 7)
            {
                //MessageBox.Show(Convert.ToString(this.typeIndex));
                String oldObj = Convert.ToString(this.bcsv.entryList[this.entryIndex].data[fieldIndex]);
                try
                {
                    switch (this.typeIndex)
                    {
                        case 0: //S32
                            this.bcsv.entryList[this.entryIndex].data[this.fieldIndex] = Convert.ToInt32(dataEntry.Text);
                            this.bcsv.entryList[this.entryIndex].dataType[this.fieldIndex] = (Byte)this.typeIndex;
                            this.bcsv.entryList[this.entryIndex].changed[this.fieldIndex] = 1; //Set the changed flag to say it's changed
                            break;
                        case 2: //Float
                            this.bcsv.entryList[this.entryIndex].data[this.fieldIndex] = Convert.ToSingle(dataEntry.Text);
                            this.bcsv.entryList[this.entryIndex].dataType[this.fieldIndex] = (Byte)this.typeIndex;
                            this.bcsv.entryList[this.entryIndex].changed[this.fieldIndex] = 1; //Set the changed flag to say it's changed
                            break;
                        case 3: //S32
                            this.bcsv.entryList[this.entryIndex].data[this.fieldIndex] = Convert.ToInt32(dataEntry.Text);
                            this.bcsv.entryList[this.entryIndex].dataType[this.fieldIndex] = (Byte)this.typeIndex;
                            this.bcsv.entryList[this.entryIndex].changed[this.fieldIndex] = 1; //Set the changed flag to say it's changed
                            break;
                        case 4: //S16
                            this.bcsv.entryList[this.entryIndex].data[this.fieldIndex] = Convert.ToInt16(dataEntry.Text);
                            this.bcsv.entryList[this.entryIndex].dataType[this.fieldIndex] = (Byte)this.typeIndex;
                            this.bcsv.entryList[this.entryIndex].changed[this.fieldIndex] = 1; //Set the changed flag to say it's changed
                            break;
                        case 5: //S8
                            this.bcsv.entryList[this.entryIndex].data[this.fieldIndex] = Convert.ToByte(dataEntry.Text);
                            this.bcsv.entryList[this.entryIndex].dataType[this.fieldIndex] = (Byte)this.typeIndex;
                            this.bcsv.entryList[this.entryIndex].changed[this.fieldIndex] = 1; //Set the changed flag to say it's changed
                            break;
                        case 6: //String
                            this.bcsv.entryList[this.entryIndex].data[this.fieldIndex] = Convert.ToString(dataEntry.Text);
                            this.bcsv.entryList[this.entryIndex].dataType[this.fieldIndex] = (Byte)this.typeIndex;
                            this.bcsv.entryList[this.entryIndex].changed[this.fieldIndex] = 1; //Set the changed flag to say it's changed
                            this.bcsv.entryList[this.entryIndex].size[this.fieldIndex] = (Byte)this.bcsv.entryList[this.entryIndex].data[this.fieldIndex].ToString().Length;
                            break;
                    }
                }
                catch (FormatException)
                {
                    if (dataEntry.Text != "")
                    {
                        dataEntry.Text = oldObj;
                    }
                }
            }

            if (this.bcsv.fieldList[fieldIndex].dataType != 1 && this.bcsv.fieldList[fieldIndex].dataType != 2 && this.bcsv.fieldList[fieldIndex].dataType != 6 && this.bcsv.fieldList[fieldIndex].dataType != 7)
            {
                try
                {
                    dataEntryShifted.Text = Convert.ToString(((Int32)this.bcsv.entryList[this.entryIndex].data[this.fieldIndex] & this.bcsv.fieldList[fieldIndex].mask) >> this.bcsv.fieldList[fieldIndex].shiftVal);
                }
                catch (InvalidCastException)
                {
                    dataEntryShifted.Text = String.Empty;
                }
            }

            else
            {
                dataEntryShifted.Text = String.Empty;
            }
        }

        private void entrySelect_SelectedIndexChanged(object sender, EventArgs e)
        {
            this.selectingEntryOrField = true;
            this.entryIndex = (UInt32)entrySelect.SelectedIndex;
            this.fieldIndex = 0;
            fieldSelect.SelectedIndex = 0;
            this.typeIndex = this.bcsv.entryList[this.entryIndex].dataType[0];
            typeSelect.SelectedIndex = this.bcsv.entryList[this.entryIndex].dataType[0];
            dataEntry.Text = Convert.ToString(this.bcsv.entryList[this.entryIndex].data[this.fieldIndex]);

            if (this.bcsv.fieldList[fieldIndex].dataType != 1 && this.bcsv.fieldList[fieldIndex].dataType != 2 && this.bcsv.fieldList[fieldIndex].dataType != 6 && this.bcsv.fieldList[fieldIndex].dataType != 7)
            {
                try
                {
                    dataEntryShifted.Text = Convert.ToString(((Int32)this.bcsv.entryList[this.entryIndex].data[this.fieldIndex] & this.bcsv.fieldList[fieldIndex].mask) >> this.bcsv.fieldList[fieldIndex].shiftVal);
                }
                catch (InvalidCastException)
                {
                    dataEntryShifted.Text = String.Empty;
                }
            }
            else
            {
                dataEntryShifted.Text = String.Empty;
            }

            //this.selectingEntryOrField = false;
        }

        private void addButton_Click(object sender, EventArgs e)
        {
            UInt32 i;

            this.bcsv.header.entryCount += 1;
            Array.Resize(ref this.bcsv.entryList, (Int32)this.bcsv.header.entryCount);

            this.bcsv.entryList[this.bcsv.header.entryCount - 1] = new BCSVEntry();
            this.bcsv.entryList[this.bcsv.header.entryCount - 1].data = new Object[this.bcsv.header.fieldCount];
            this.bcsv.entryList[this.bcsv.header.entryCount - 1].dataType = new Byte[this.bcsv.header.fieldCount];
            this.bcsv.entryList[this.bcsv.header.entryCount - 1].changed = new Byte[this.bcsv.header.fieldCount];
            this.bcsv.entryList[this.bcsv.header.entryCount - 1].offset = new UInt32[this.bcsv.header.fieldCount];
            this.bcsv.entryList[this.bcsv.header.entryCount - 1].size = new Byte[this.bcsv.header.fieldCount];
            this.bcsv.entryList[this.bcsv.header.entryCount - 1].stringOffs = new UInt32[this.bcsv.header.fieldCount];

            for (i = 0; i < this.bcsv.header.fieldCount; i++)
            {
                this.bcsv.entryList[this.bcsv.header.entryCount - 1].offset[i] = ((this.bcsv.header.entryCount - 1) * this.bcsv.header.entryDataSize) + this.bcsv.fieldList[i].dataOffset;
                this.bcsv.entryList[this.bcsv.header.entryCount - 1].dataType[i] = this.bcsv.fieldList[i].dataType;
                this.bcsv.entryList[this.bcsv.header.entryCount - 1].changed[i] = 0;
                if (this.bcsv.fieldList[i].dataType == 0 || this.bcsv.fieldList[i].dataType == 3) //S32
                {
                    this.bcsv.entryList[this.bcsv.header.entryCount - 1].data[i] = (Int32)0;
                    this.bcsv.entryList[this.bcsv.header.entryCount - 1].size[i] = 4;
                }
                else if (this.bcsv.fieldList[i].dataType == 2) //Float
                {
                    this.bcsv.entryList[this.bcsv.header.entryCount - 1].data[i] = (Single)0;
                    this.bcsv.entryList[this.bcsv.header.entryCount - 1].size[i] = 4;
                }
                else if (this.bcsv.fieldList[i].dataType == 4) //S16
                {
                    this.bcsv.entryList[this.bcsv.header.entryCount - 1].data[i] = (Int16)0;
                    this.bcsv.entryList[this.bcsv.header.entryCount - 1].size[i] = 2;
                }
                else if (this.bcsv.fieldList[i].dataType == 5) //S8
                {
                    this.bcsv.entryList[this.bcsv.header.entryCount - 1].data[i] = (Byte)0;
                    this.bcsv.entryList[this.bcsv.header.entryCount - 1].size[i] = 1;
                }
                else if (this.bcsv.fieldList[i].dataType == 6) //String
                {
                    this.bcsv.entryList[this.bcsv.header.entryCount - 1].data[i] = "Placeholder";
                    this.bcsv.entryList[this.bcsv.header.entryCount - 1].size[i] = 11;
                    this.bcsv.entryList[this.bcsv.header.entryCount - 1].stringOffs[i] = this.bcsv.entryList[this.bcsv.oldI].stringOffs[this.bcsv.oldJ] + (UInt32)12;
                }
            }

            entrySelect.Items.Add("Entry " + Convert.ToString(this.bcsv.header.entryCount - 1));
        }

        private void removeButton_Click(object sender, EventArgs e)
        {
            /* this.bcsv.header.entryCount -= 1;

            var tmp = this.bcsv.entryList.ToList();
            tmp.RemoveAt((Int32)this.entryIndex);
            this.bcsv.entryList = tmp.ToArray();

            entrySelect.Items.RemoveAt((Int32)this.entryIndex);

            this.entryIndex = this.bcsv.header.entryCount - 1;
            entrySelect.SelectedIndex = (Int32)this.bcsv.header.entryCount - 1; */
        }

        private void duplicateButton_Click(object sender, EventArgs e)
        {
            try
            {
                this.bcsv.header.entryCount += 1;
                Array.Resize(ref this.bcsv.entryList, (Int32)this.bcsv.header.entryCount);

                this.bcsv.entryList[this.bcsv.header.entryCount - 1] = this.bcsv.entryList[entrySelect.SelectedIndex];
                entrySelect.Items.Add("Entry " + Convert.ToString(this.bcsv.header.entryCount - 1));
            }
            catch (IndexOutOfRangeException ex)
            {
                this.bcsv.header.entryCount -= 1;
                Array.Resize(ref this.bcsv.entryList, (Int32)this.bcsv.header.entryCount);
            }
        }
    }
}
